from orchestra.bots.basebot import BaseBot
from orchestra.bots.errors import SlackCommandInvalidRequest
from orchestra.bots.tests.fixtures import get_mock_slack_data
from orchestra.tests.helpers import OrchestraTestCase


class BaseBotTest(OrchestraTestCase):
    token = get_mock_slack_data().get('token')

    def test_help(self):
        bot = BaseBot(self.token)
        mock_slack_data = get_mock_slack_data(text='help')
        with self.assertRaises(NotImplementedError):
            bot.dispatch(mock_slack_data)

    def test_validate(self):
        """
            Ensure we only listen to valid requests.
        """
        mock_slack_data = get_mock_slack_data()
        # Test all requests allowed
        bot = BaseBot(self.token)
        self.assertEqual(mock_slack_data, bot.validate(mock_slack_data))

        # verify we validate the token
        bot = BaseBot('')
        with self.assertRaises(SlackCommandInvalidRequest):
            bot.validate(mock_slack_data)

        # verify that we perform validation on each of the fields
        validated_fields = ['allowed_team_ids', 'allowed_domains',
                            'allowed_channel_ids', 'allowed_channel_names',
                            'allowed_user_ids', 'allowed_user_names',
                            'allowed_commands']
        for field in validated_fields:
            config = {field: []}
            bot = BaseBot(self.token, **config)
            with self.assertRaises(SlackCommandInvalidRequest):
                bot.validate(mock_slack_data)
        config = {'allowed_{}s'.format(field): [mock_slack_data.get(field)]
                  for field in validated_fields}
        bot = BaseBot(self.token, **config)
        self.assertEqual(mock_slack_data, bot.validate(mock_slack_data))

    def test_dispatch(self):
        bot = BaseBot(self.token)
        bot.commands = (
            (r'test_command (?P<test_param>[0-9]+)', 'test_command'),
        )

        def test_command(test_param):
            return test_param
        bot.test_command = test_command

        # Assign the testing command
        bot.__init__(self.token)

        # Test a valid command
        text = 'test_command 5'
        mock_slack_data = get_mock_slack_data(text=text)
        result = bot.dispatch(mock_slack_data)
        self.assertEqual('5', result)

        # Test a valid command with missing param
        text = 'test_command'
        mock_slack_data = get_mock_slack_data(text=text)
        result = bot.dispatch(mock_slack_data)
        self.assertEqual(bot.no_command_found(text), result)

        # Test invalid command
        text = 'invalid'
        mock_slack_data = get_mock_slack_data(text=text)
        result = bot.dispatch(mock_slack_data)
        self.assertEqual(bot.no_command_found(text), result)



from datetime import timedelta
from unittest.mock import patch

import factory
from dateutil.parser import parse
from django.contrib.auth.models import User
from django.test import Client
from django.test import override_settings
from django.utils import timezone

from orchestra.communication.slack import create_project_slack_group
from orchestra.models import CommunicationPreference
from orchestra.models import Iteration
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.models import Step
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import WorkerCertification
from orchestra.models import Workflow
from orchestra.models import WorkflowVersion
from orchestra.tests.helpers.iterations import verify_iterations
from orchestra.tests.helpers.workflow import workflow_fixtures
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import submit_task
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import current_assignment
from orchestra.utils.task_properties import get_iteration_history
from orchestra.workflow.defaults import get_default_assignment_policy
from orchestra.workflow.defaults import get_default_creation_policy
from orchestra.workflow.defaults import get_default_review_policy

BASE_DATETIME = parse('2015-10-12T00:00:00+00:00')
ITERATION_DURATION = timedelta(hours=1)
PICKUP_DELAY = timedelta(hours=1)


def get_detailed_description(task_details, text=None):
    if text is None:
        text = 'No text given'
    return '{} {}'.format(text, task_details['step']['slug'])


class WorkflowFactory(factory.django.DjangoModelFactory):

    code_directory = factory.Sequence(
        lambda n: 'Code Directory {}'.format(n))
    slug = factory.Sequence(
        lambda n: 'Slug {}'.format(n))

    class Meta:
        model = Workflow


class WorkflowVersionFactory(factory.django.DjangoModelFactory):

    workflow = factory.SubFactory(WorkflowFactory)

    class Meta:
        model = WorkflowVersion


class StepFactory(factory.django.DjangoModelFactory):

    is_human = False
    workflow_version = factory.SubFactory(WorkflowVersionFactory)

    class Meta:
        model = Step


class UserFactory(factory.django.DjangoModelFactory):
    username = factory.Sequence(
        lambda n: 'Username {}'.format(n))

    class Meta:
        model = User

    password = factory.PostGenerationMethodCall('set_password',
                                                'defaultpassword')


class WorkerFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = 'orchestra.Worker'


class CertificationFactory(factory.django.DjangoModelFactory):

    slug = factory.Sequence(
        lambda n: 'Slug {}'.format(n))
    name = factory.Sequence(
        lambda n: 'Name {}'.format(n))
    description = factory.Sequence(
        lambda n: 'Description {}'.format(n))
    workflow = factory.SubFactory(WorkflowFactory)

    class Meta:
        model = 'orchestra.Certification'


class WorkerCertificationFactory(factory.django.DjangoModelFactory):
    certification = factory.SubFactory(CertificationFactory)
    worker = factory.SubFactory(WorkerFactory)
    task_class = WorkerCertification.TaskClass.REAL
    role = WorkerCertification.Role.ENTRY_LEVEL
    staffing_priority = 0

    class Meta:
        model = 'orchestra.WorkerCertification'


class PayRateFactory(factory.django.DjangoModelFactory):
    worker = factory.SubFactory(WorkerFactory)
    hourly_multiplier = 1

    class Meta:
        model = 'orchestra.PayRate'


class ProjectFactory(factory.django.DjangoModelFactory):

    class Meta:
        model = 'orchestra.Project'

    workflow_version = factory.SubFactory(WorkflowVersionFactory)
    short_description = factory.Sequence(
        lambda n: 'Project {}'.format(n))
    priority = 0
    task_class = WorkerCertification.TaskClass.REAL


class TaskFactory(factory.django.DjangoModelFactory):

    step = factory.SubFactory(StepFactory)
    project = factory.SubFactory(ProjectFactory)
    status = Task.Status.AWAITING_PROCESSING

    class Meta:
        model = 'orchestra.Task'


class TaskAssignmentFactory(factory.django.DjangoModelFactory):
    task = factory.SubFactory(TaskFactory)
    status = TaskAssignment.Status.PROCESSING

    class Meta:
        model = 'orchestra.TaskAssignment'


class TimeEntryFactory(factory.django.DjangoModelFactory):

    class Meta:
        model = 'orchestra.TimeEntry'


class TodoFactory(factory.django.DjangoModelFactory):
    task = factory.SubFactory(TaskFactory)
    completed = False
    description = factory.Sequence(
        lambda n: 'Description {}'.format(n))
    start_by_datetime = None
    due_datetime = None

    class Meta:
        model = 'orchestra.Todo'


class TodoQAFactory(factory.django.DjangoModelFactory):
    todo = factory.SubFactory(TodoFactory)
    approved = True
    comment = factory.Sequence(
        lambda n: 'Comment {}'.format(n))

    class Meta:
        model = 'orchestra.TodoQA'


class TodoListTemplateFactory(factory.django.DjangoModelFactory):
    slug = factory.Sequence(
        lambda n: 'todolist_template_{}'.format(n))
    name = factory.Sequence(
        lambda n: 'TodoList template {}'.format(n))
    description = factory.Sequence(
        lambda n: 'Description {}'.format(n))
    creator = None
    todos = {'items': []}

    class Meta:
        model = 'orchestra.TodoListTemplate'


class CommunicationPreferenceFactory(factory.django.DjangoModelFactory):

    worker = factory.SubFactory(WorkerFactory)
    communication_type = (CommunicationPreference.CommunicationType.
                          TASK_STATUS_CHANGE.value)
    methods = CommunicationPreference.get_default_methods()

    class Meta:
        model = 'orchestra.CommunicationPreference'


class StaffBotRequestFactory(factory.django.DjangoModelFactory):

    task = factory.SubFactory(TaskFactory)
    request_cause = StaffBotRequest.RequestCause.USER.value
    required_role_counter = 0

    class Meta:
        model = StaffBotRequest


class StaffingRequestInquiryFactory(factory.django.DjangoModelFactory):

    communication_preference = factory.SubFactory(
        CommunicationPreferenceFactory)
    request = factory.SubFactory(StaffBotRequestFactory)
    communication_method = (
        StaffingRequestInquiry.CommunicationMethod.SLACK.value)

    class Meta:
        model = StaffingRequestInquiry


class StaffingResponseFactory(factory.django.DjangoModelFactory):

    request_inquiry = factory.SubFactory(StaffingRequestInquiryFactory)
    is_available = False

    class Meta:
        model = StaffingResponse


@override_settings(ORCHESTRA_SLACK_EXPERTS_ENABLED=True)
def setup_models(test_case):
    """ Set up models that we'll use in multiple tests """

    # Workflow generation data
    workflows = workflow_fixtures

    # Worker generation data
    workers = {
        0: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
        ],
        1: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.REVIEWER),
        ],
        2: [],
        3: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.REVIEWER),
        ],
        4: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2_ap', WorkerCertification.Role.ENTRY_LEVEL),
        ],
        5: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2_ap', WorkerCertification.Role.REVIEWER),
        ],
        6: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1', WorkerCertification.Role.REVIEWER),
            ('certification2', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.REVIEWER),
            ('certification2_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification2_ap', WorkerCertification.Role.REVIEWER),
        ],
        7: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.REVIEWER),
        ],
        8: [
            ('certification1', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1', WorkerCertification.Role.REVIEWER),
            ('certification1_ap', WorkerCertification.Role.ENTRY_LEVEL),
            ('certification1_ap', WorkerCertification.Role.REVIEWER),
        ]
    }

    # Project generation data
    projects = {
        'empty_project': 'test_workflow',
        'base_test_project': 'test_workflow',
        'test_human_and_machine': 'test_workflow_2',
        'no_task_assignments': 'test_workflow',
        'reject_entry_proj': 'test_workflow',
        'reject_rev_proj': 'test_workflow',
        'aborted_project': 'test_workflow',
        'project_to_end': 'test_workflow',
        'assignment_policy': 'assignment_policy_workflow',
        'project_management_project': 'test_workflow',
        'staffbot_assignment_policy': 'staffbot_assignment_policy_workflow',
        'creation_policy': 'creation_policy_workflow',
        'next_todo_proj': 'test_workflow',
        'sanitybot': 'sanitybot_workflow',
    }

    # Task generation data
    test_case.test_version_slug = 'test_workflow'
    test_case.test_step_slug = 'step1'
    test_data = {'test_key': 'test_value'}
    tasks = {
        'review_task': {
            'project_name': 'base_test_project',
            'status': Task.Status.PENDING_REVIEW,
            'assignments': [
                (0, test_data, TaskAssignment.Status.SUBMITTED)
            ],
        },
        'awaiting_processing': {
            'project_name': 'no_task_assignments',
            'status': Task.Status.AWAITING_PROCESSING,
            'assignments': []
        },
        'rejected_entry': {
            'project_name': 'reject_entry_proj',
            'status': Task.Status.POST_REVIEW_PROCESSING,
            'assignments': [
                (4, test_data, TaskAssignment.Status.PROCESSING),
                (6, {}, TaskAssignment.Status.SUBMITTED)
            ]
        },
        'rejected_review': {
            'project_name': 'reject_rev_proj',
            'status': Task.Status.POST_REVIEW_PROCESSING,
            'assignments': [
                (5, {}, TaskAssignment.Status.SUBMITTED),
                (6, test_data, TaskAssignment.Status.PROCESSING),
                (7, {}, TaskAssignment.Status.SUBMITTED)
            ]
        },
        'project_management_task': {
            'project_name': 'project_management_project',
            'status': Task.Status.POST_REVIEW_PROCESSING,
            'assignments': [
                (5, {}, TaskAssignment.Status.SUBMITTED),
                (6, test_data, TaskAssignment.Status.PROCESSING),
                (7, {}, TaskAssignment.Status.SUBMITTED)
            ]
        },
        'aborted': {
            'project_name': 'aborted_project',
            'status': Task.Status.ABORTED,
            'assignments': [
                (4, test_data, TaskAssignment.Status.PROCESSING)
            ]
        },
        'next_todo_task': {
            'project_name': 'next_todo_proj',
            'status': Task.Status.PROCESSING,
            'assignments': [
                (5, {}, TaskAssignment.Status.PROCESSING)
            ]
        },
    }

    # Create the objects
    _setup_workflows(test_case, workflows)
    _setup_workers(test_case, workers)
    _setup_projects(test_case, projects)
    _setup_tasks(test_case, tasks)


def _setup_workflows(test_case, workflows):
    # Create workflows
    test_case.workflows = {}
    test_case.certifications = {}
    test_case.workflow_versions = {}
    test_case.workflow_steps = {}
    for workflow_idx, workflow_details in enumerate(workflows):
        workflow = WorkflowFactory(
            slug=workflow_details['slug'],
            name=workflow_details['name'],
            code_directory='/test/dir/{}'.format(workflow_idx),
        )
        test_case.workflows[workflow_details['slug']] = workflow

        # Create certifications and dependencies
        # TODO(joshblum): workflow_step slugs may not be unique across
        # workflows! we shouldn't depend on this here
        for cert_details in workflow_details.get('certifications', []):
            certification = CertificationFactory(
                slug=cert_details['slug'],
                name=cert_details['name'],
                workflow=workflow,
            )
            test_case.certifications[cert_details['slug']] = certification

            for required_slug in cert_details['required_certifications']:
                certification.required_certifications.add(
                    test_case.certifications[required_slug])

        # Create workflow versions
        for version_details in workflow_details['versions']:
            version = WorkflowVersionFactory(
                workflow=workflow,
                slug=version_details['slug'],
                name=version_details['name'],
                description=version_details['description'],
                sanity_checks=version_details.get('sanity_checks', {})
            )
            test_case.workflow_versions[version_details['slug']] = version

            # Create workflow steps
            # TODO(joshblum): workflow_step slugs may not be unique across
            # workflows! we shouldn't depend on this here
            test_case.workflow_steps[version.slug] = {}
            workflow_steps = test_case.workflow_steps[version.slug]
            workflow_step_backrefs = []
            for step_details in version_details['steps']:
                is_human = step_details['is_human']
                step = StepFactory(
                    workflow_version=version,
                    slug=step_details['slug'],
                    name=step_details['name'],
                    is_human=is_human,
                    description=step_details.get('description', ''),

                    # TODO(dhaas): make default policies work
                    review_policy=step_details.get(
                        'review_policy',
                        get_default_review_policy(is_human)
                    ),
                    assignment_policy=step_details.get(
                        'assignment_policy',
                        get_default_assignment_policy(is_human)
                    ),
                    creation_policy=step_details.get(
                        'creation_policy',
                        get_default_creation_policy()),
                    user_interface=step_details.get('user_interface', {}),
                    execution_function=step_details.get('execution_function',
                                                        {}),
                )
                workflow_steps[step.slug] = step

                # Add required certifications
                for required_slug in step_details.get(
                        'required_certifications', []):
                    step.required_certifications.add(
                        test_case.certifications[required_slug])

                # Add step dependencies
                workflow_step_backrefs.extend(
                    _add_step_dependencies(
                        step_details, workflow_steps, 'creation_depends_on'))
                workflow_step_backrefs.extend(
                    _add_step_dependencies(
                        step_details, workflow_steps, 'submission_depends_on'))

            # Create backreferences we missed
            _create_backrefs(test_case.workflow_steps[version.slug],
                             workflow_step_backrefs)


def _add_step_dependencies(step_dict, existing_workflow_steps, attr):
    backrefs = []
    step = existing_workflow_steps.get(step_dict['slug'])
    for dependent_slug in step_dict.get(attr, []):
        dependent_step = existing_workflow_steps.get(dependent_slug)

        # Some workflow fixtures have intentional step cycles, so
        # we create the backward references later.
        if not dependent_step:
            backrefs.append({
                'attr': attr,
                'step': step,
                'ref': dependent_slug,
            })
            continue

        getattr(step, attr).add(dependent_step)
    return backrefs


def _create_backrefs(workflow_steps, backrefs):
    for backref in backrefs:
        dependency = workflow_steps[backref['ref']]
        getattr(backref['step'], backref['attr']).add(dependency)


def _setup_workers(test_case, workers):
    # Create and certify workers
    test_case.workers = {}
    test_case.clients = {}
    test_case.comm_prefs = {}
    for user_id, certifications in workers.items():
        # Create user, worker, client
        user = (UserFactory(username='test_user_{}'.format(user_id),
                            first_name='test_first_{}'.format(user_id),
                            last_name='test_last_{}'.format(user_id),
                            email='test_user_{}@test.com'.format(user_id)))
        worker = WorkerFactory(user=user,
                               slack_user_id='test_user_{}'.format(user_id))
        test_case.workers[user_id] = worker
        test_case.comm_prefs[
            user_id] = (CommunicationPreference.objects.
                        get_or_create_all_types(worker))
        test_case.clients[user_id] = Client()
        test_case.clients[user_id].login(
            username='test_user_{}'.format(user_id),
            password='defaultpassword')

        # Assign certifications
        for slug, role in certifications:
            WorkerCertificationFactory(
                certification=test_case.certifications[slug],
                worker=test_case.workers[user_id],
                role=role)


def _setup_projects(test_case, projects):
    # Create projects
    test_case.projects = {}
    for name, workflow_slug in projects.items():
        test_case.projects[name] = ProjectFactory(
            start_datetime=BASE_DATETIME,
            workflow_version=test_case.workflow_versions[workflow_slug])
        create_project_slack_group(test_case.projects[name])


def _setup_tasks(test_case, tasks):
    # Create and assign tasks
    test_case.tasks = {}
    test_case.test_step = test_case.workflow_steps[
        test_case.test_version_slug][test_case.test_step_slug]
    for task_slug, details in tasks.items():
        task_pickup_time = BASE_DATETIME + timedelta(hours=1)
        task = TaskFactory(
            project=test_case.projects[details['project_name']],
            step=test_case.test_step,
            status=details['status'],
            start_datetime=task_pickup_time,
        )
        test_case.tasks[task_slug] = task
        for i, (user_id,
                task_data,
                assignment_status) in enumerate(details['assignments']):
            assignment = TaskAssignmentFactory(
                worker=test_case.workers[user_id],
                task=task,
                status=assignment_status,
                assignment_counter=i,
                in_progress_task_data=task_data,
                start_datetime=_new_assignment_start_datetime(task))

            # Each assignment must have at least one corresponding iteration
            Iteration.objects.create(
                assignment=assignment,
                start_datetime=assignment.start_datetime,
                end_datetime=assignment.start_datetime + ITERATION_DURATION,
                submitted_data=assignment.in_progress_task_data,
                status=Iteration.Status.REQUESTED_REVIEW)

            # Create time entry for each task.
            TimeEntryFactory(date='2016-04-04',
                             time_worked=timedelta(minutes=30),
                             assignment=assignment,
                             worker=test_case.workers[user_id],
                             description=(
                                 'test description {}'.format(assignment.id)))

        cur_assignment = current_assignment(task)
        assignments = assignment_history(task).all()
        if cur_assignment and (
                cur_assignment.status == TaskAssignment.Status.PROCESSING):
            # If there's a currently processing assignment, we'll need to
            # adjust the task's iteration sequence
            processing_counter = cur_assignment.assignment_counter
            if processing_counter != len(assignments) - 1:
                # If processing assignment is not the last in the hierarchy, we
                # need to reconstruct an iteration sequence: REQUESTED_REVIEW
                # up to the highest assignment counter, then PROVIDED_REVIEW
                # back down to the current assignment
                last_iteration = assignments.last().iterations.first()
                last_iteration.status = Iteration.Status.PROVIDED_REVIEW
                last_iteration.save()

                adjust_assignments = list(assignments)[processing_counter:-1]
                for assignment in reversed(adjust_assignments):
                    last_iteration = get_iteration_history(task).last()
                    Iteration.objects.create(
                        assignment=assignment,
                        start_datetime=last_iteration.end_datetime,
                        end_datetime=(
                            last_iteration.end_datetime + ITERATION_DURATION),
                        submitted_data=assignment.in_progress_task_data,
                        status=Iteration.Status.PROVIDED_REVIEW)

            # If there is a currently processing assignment, the task's last
            # iteration should still be processing
            last_iteration = get_iteration_history(task).last()
            last_iteration.end_datetime = None
            last_iteration.submitted_data = {}
            last_iteration.status = Iteration.Status.PROCESSING
            last_iteration.save()

        verify_iterations(task.id)


def _new_assignment_start_datetime(task):
    # Each assignment is created after the previous assignment's first
    # iteration is complete and a pickup delay elapses
    previous_assignment_duration = (
        task.assignments.count() * (ITERATION_DURATION + PICKUP_DELAY))
    return task.start_datetime + previous_assignment_duration + PICKUP_DELAY


def setup_complete_task(test_case):
    # Microseconds are truncated when manually saving models
    test_start = timezone.now().replace(microsecond=0)
    times = {
        'awaiting_pickup': test_start,
        'entry_pickup': test_start + timedelta(hours=1),
        'entry_submit': test_start + timedelta(hours=2),
        'reviewer_pickup': test_start + timedelta(hours=3),
        'reviewer_reject': test_start + timedelta(hours=4),
        'entry_resubmit': test_start + timedelta(hours=5),
        'reviewer_accept': test_start + timedelta(hours=6),
    }

    task = TaskFactory(
        project=test_case.projects['empty_project'],
        status=Task.Status.AWAITING_PROCESSING,
        step=test_case.test_step,
        start_datetime=times['awaiting_pickup'])

    workers = {
        'entry': test_case.workers[0],
        'reviewer': test_case.workers[1]
    }

    assign_task(workers['entry'].id, task.id)

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.PROCESSING)

    submit_task(
        task.id, {'test': 'entry_submit'},
        Iteration.Status.REQUESTED_REVIEW,
        workers['entry'])

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.PENDING_REVIEW)

    assign_task(workers['reviewer'].id, task.id)
    reviewer_assignment = task.assignments.get(
        worker=workers['reviewer'])

    # Modify assignment with correct datetime
    reviewer_assignment.start_datetime = times['reviewer_pickup']
    reviewer_assignment.save()

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.REVIEWING)

    submit_task(
        task.id, {'test': 'reviewer_reject'},
        Iteration.Status.PROVIDED_REVIEW,
        workers['reviewer'])

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.POST_REVIEW_PROCESSING)

    submit_task(
        task.id, {'test': 'entry_resubmit'},
        Iteration.Status.REQUESTED_REVIEW,
        workers['entry'])

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.REVIEWING)

    with patch('orchestra.utils.task_lifecycle._is_review_needed',
               return_value=False):
        submit_task(
            task.id, {'test': 'reviewer_accept'},
            Iteration.Status.REQUESTED_REVIEW,
            workers['reviewer'])

    task.refresh_from_db()
    test_case.assertEqual(task.status, Task.Status.COMPLETE)
    test_case.assertEqual(task.assignments.count(), 2)
    for assignment in task.assignments.all():
        test_case.assertEqual(
            assignment.status, TaskAssignment.Status.SUBMITTED)
        test_case.assertEqual(assignment.iterations.count(), 2)

    # Modify assignments with correct datetime
    new_datetime_labels = ('entry_pickup', 'reviewer_pickup')
    for i, assignment in enumerate(assignment_history(task).all()):
        assignment.start_datetime = times[new_datetime_labels[i]]
        assignment.save()

    # Modify iterations with correct datetime
    new_datetime_labels = (
        ('entry_pickup', 'entry_submit'),
        ('reviewer_pickup', 'reviewer_reject'),
        ('reviewer_reject', 'entry_resubmit'),
        ('entry_resubmit', 'reviewer_accept')
    )
    new_datetimes = [
        (times[start_label], times[end_label])
        for start_label, end_label in new_datetime_labels]

    for i, iteration in enumerate(get_iteration_history(task)):
        iteration.start_datetime, iteration.end_datetime = new_datetimes[i]
        iteration.save()

    verify_iterations(task.id)

    return task



import json

from django.utils import timezone
from dateutil.parser import parse
from django.core.urlresolvers import reverse

from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.models import Worker
from orchestra.project_api.serializers import TimeEntrySerializer
from orchestra.tests.helpers import EndpointTestCase
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import TodoQAFactory
from orchestra.tests.helpers.fixtures import TodoListTemplateFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.todos.serializers import TodoSerializer
from orchestra.todos.serializers import TodoQASerializer
from orchestra.todos.serializers import TodoListTemplateSerializer
from orchestra.utils.load_json import load_encoded_json


def _todo_data(task, description, completed,
               skipped_datetime=None, start_by=None,
               due=None, parent_todo=None, template=None,
               activity_log=str({'actions': []}), qa=None):
    return {
        'task': task.id,
        'completed': completed,
        'description': description,
        'template': template,
        'parent_todo': parent_todo,
        'start_by_datetime': start_by,
        'due_datetime': due,
        'activity_log': activity_log,
        'skipped_datetime': skipped_datetime,
        'qa': qa
    }


def _get_test_conditional_props(project):
    return {
        'prop1': True,
        'prop2': False
    }


class TodosEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.list_create_url = reverse('orchestra:todos:todos')
        self.list_details_url_name = 'orchestra:todos:todo'
        self.tasks = Task.objects.filter(assignments__worker=self.worker)
        self.task = self.tasks[0]
        self.todo_description = 'Let us do this'
        self.deadline = parse('2018-01-16T07:03:00+00:00')

    def _verify_missing_task(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'], 'No task for given id')
        self.assertEqual(data['error'], 400)

    def _verify_worker_not_assigned(self, response):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'],
                         'Worker is not assigned to this task id.')
        self.assertEqual(data['error'], 400)

    def _verify_time_entries(self, data):
        for time_entry in data:
            serializer = TimeEntrySerializer(data=time_entry)
            self.assertTrue(serializer.is_valid())

    def _verify_todo_content(self, todo, expected_todo):
        todo = dict(todo)
        created_at = todo.pop('created_at')
        todo_id = todo.pop('id')
        self.assertEqual(todo, expected_todo)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_id, 0)

    def _verify_todos_list(self, project_id, expected_todos, success):
        resp = self.request_client.get(self.list_create_url,
                                       {'project': project_id})
        if success:
            self.assertEqual(resp.status_code, 200)
            data = load_encoded_json(resp.content)
            for todo, expected_todo in zip(data, expected_todos):
                self._verify_todo_content(todo, expected_todo)
        else:
            self.assertEqual(resp.status_code, 403)

    def _verify_todo_creation(self, task, success):
        num_todos = Todo.objects.all().count()
        resp = self.request_client.post(self.list_create_url, {
            'task': task.id,
            'description': self.todo_description})
        if success:
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(Todo.objects.all().count(), num_todos + 1)
            todo = load_encoded_json(resp.content)
            self._verify_todo_content(
                todo, _todo_data(task, self.todo_description, False))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertEqual(Todo.objects.all().count(), num_todos)

    def _verify_todo_update(self, todo, success):
        description = 'new description'
        list_details_url = reverse(
            self.list_details_url_name,
            kwargs={'pk': todo.id})
        resp = self.request_client.put(
            list_details_url,
            json.dumps(_todo_data(todo.task, description, True)),
            content_type='application/json')
        updated_todo = TodoSerializer(Todo.objects.get(id=todo.id)).data
        if success:
            self.assertEqual(resp.status_code, 200)
            self._verify_todo_content(
                updated_todo, _todo_data(todo.task, description, True))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertNotEqual(updated_todo['description'], description)

    def test_todos_list_create(self):
        self._verify_todos_list(self.task.project.id, [], True)
        self._verify_todo_creation(self.task, True)
        self._verify_todos_list(self.task.project.id,
                                [_todo_data(
                                    self.task, self.todo_description, False)],
                                True)

    def test_todos_list_create_permissions(self):
        # Can't make requests for projects in which you're uninvolved.
        task = TaskFactory()
        self._verify_todos_list(task.project.id, [], False)
        self._verify_todo_creation(task, False)

    def test_todo_details_and_permissions(self):
        # You should be able to update Todos for projects in which
        # you're involved, and not for other projects.
        good_todo = TodoFactory(task=self.task)
        self._verify_todo_update(good_todo, True)
        bad_todo = TodoFactory()
        self._verify_todo_update(bad_todo, False)

    def test_create_todo_with_start_by_datetime(self):
        START_DESCRIPTION = 'Start soon'

        start_by_todo = TodoFactory(
            task=self.task,
            start_by_datetime=self.deadline,
            description=START_DESCRIPTION)

        self._verify_todos_list(self.task.project.id, [
            _todo_data(
                start_by_todo.task,
                START_DESCRIPTION,
                False,
                None,
                self.deadline.strftime('%Y-%m-%dT%H:%M:%SZ'),
                None)
        ], True)

    def test_create_todo_with_due_datetime(self):
        DUE_DESCRIPTION = 'Due soon'

        due_todo = TodoFactory(
            task=self.task,
            due_datetime=self.deadline,
            description=DUE_DESCRIPTION)

        self._verify_todos_list(self.task.project.id, [
            _todo_data(
                due_todo.task,
                DUE_DESCRIPTION,
                False,
                None,
                None,
                self.deadline.strftime('%Y-%m-%dT%H:%M:%SZ')),
        ], True)


class TodoQAEndpointTests(EndpointTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.list_create_url = reverse('orchestra:todos:todo_qas')
        self.worker_task_recent_todo_qas_url = reverse(
            'orchestra:todos:worker_task_recent_todo_qas')
        self.list_details_url_name = 'orchestra:todos:todo_qa'
        self.tasks = Task.objects.filter(
            assignments__worker=self.worker)
        self.task_0 = self.tasks[0]
        self.task_1 = self.tasks[1]
        self.todo = TodoFactory(task=self.task_0)
        self.comment = 'Test comment'

    def _todo_qa_data(
            self, todo, approved, comment):
        return {
            'todo': todo.id,
            'approved': approved,
            'comment': comment
        }

    def _verify_todo_qa_content(self, todo_qa,
                                expected_todo_qa):
        todo_qa = dict(todo_qa)
        created_at = todo_qa.pop('created_at')
        todo_qa_id = todo_qa.pop('id')
        self.assertEqual(todo_qa, expected_todo_qa)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_qa_id, 0)

    def _verify_todo_qa_creation(self, todo, success):
        num_todo_qas = TodoQA.objects.all().count()
        resp = self.request_client.post(self.list_create_url, {
            'todo': todo.id,
            'approved': True,
            'comment': self.comment})
        if success:
            self.assertEqual(resp.status_code, 201)
            self.assertEqual(TodoQA.objects.all().count(), num_todo_qas + 1)
            todo_qa = load_encoded_json(resp.content)
            self._verify_todo_qa_content(todo_qa, self._todo_qa_data(
                todo, True, self.comment))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertEqual(TodoQA.objects.all().count(), num_todo_qas)

    def _verify_todo_qa_update(self, todo_qa, success):
        comment = 'new comment'
        list_details_url = reverse(
            self.list_details_url_name,
            kwargs={'pk': todo_qa.id})
        resp = self.request_client.put(
            list_details_url,
            json.dumps(self._todo_qa_data(
                todo_qa.todo, True, comment)),
            content_type='application/json')
        updated_todo_qa = TodoQASerializer(
            TodoQA.objects.get(id=todo_qa.id)).data
        if success:
            self.assertEqual(resp.status_code, 200)
            self._verify_todo_qa_content(updated_todo_qa, self._todo_qa_data(
                todo_qa.todo, True, comment))
        else:
            self.assertEqual(resp.status_code, 403)
            self.assertNotEqual(
                updated_todo_qa['comment'], comment)

    def test_todo_qas_list_create(self):
        self._verify_todo_qa_creation(self.todo, True)

    def test_todo_qas_create_permissions(self):
        # Can't make requests for projects in which you're uninvolved.
        todo = TodoFactory()
        self._verify_todo_qa_creation(todo, False)

    def test_todo_qa_details_and_permissions(self):
        # You should be able to update TodoQAs for projects in which
        # you're involved, and not for other projects.
        good_todo_qa = TodoQAFactory(todo=self.todo)
        self._verify_todo_qa_update(good_todo_qa, True)
        bad_todo_qa = TodoQAFactory()
        self._verify_todo_qa_update(bad_todo_qa, False)

    def _verify_worker_task_recent_todo_qas(self, task, todo_qa, success):
        resp = self.request_client.get(self.worker_task_recent_todo_qas_url,
                                       {'task': task.id})
        if success:
            self.assertEqual(resp.status_code, 200)
            data = load_encoded_json(resp.content)
            if todo_qa:
                self.assertEqual(
                    TodoQASerializer(todo_qa).data,
                    list(data.values())[0])
                self.assertEqual(1, len(data.keys()))
            else:
                self.assertEqual({}, data)
        else:
            self.assertEqual(resp.status_code, 403)

    def test_worker_task_recent_todo_qas(self):
        todo_task_0 = TodoFactory(task=self.task_0)
        todo_task_1 = TodoFactory(task=self.task_1)

        # Zero TodoQAs
        self._verify_worker_task_recent_todo_qas(
            self.task_0, None, True)

        todo_qa_task_0 = TodoQAFactory(todo=todo_task_0, approved=False)

        # Most recent TodoQA is todo_qa_task_0
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_0, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_0, True)

        todo_qa_task_1 = TodoQAFactory(todo=todo_task_1, approved=False)

        # If available use the todo qa for the corresponding task.
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_0, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_1, True)

        todo_qa_task_0.delete()

        # Most recent TodoQA is todo_qa_task_1
        self._verify_worker_task_recent_todo_qas(
            self.task_0, todo_qa_task_1, True)

        self._verify_worker_task_recent_todo_qas(
            self.task_1, todo_qa_task_1, True)

        # Can't make requests for projects in which you're uninvolved.
        bad_task = TaskFactory()
        todo_bad_task = TodoFactory(task=bad_task)
        todo_qa_bad_task = TodoQAFactory(todo=todo_bad_task, approved=False)
        self._verify_worker_task_recent_todo_qas(
            bad_task, todo_qa_bad_task, False)


class TodoTemplateEndpointTests(EndpointTestCase):
    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.request_client.login(username=self.worker.user.username,
                                  password='defaultpassword')
        self.todolist_template_list_url = reverse(
            'orchestra:todos:todolist_templates')
        self.todolist_template_detail_url_name = \
            'orchestra:todos:todolist_template'
        self.todolist_template_slug = 'test_todolist_template_slug'
        self.todolist_template_name = 'test_todolist_template_name'
        self.todolist_template_description = \
            'test_todolist_template_description'
        self.tasks = Task.objects.filter(assignments__worker=self.worker)
        self.task = self.tasks[0]

    def _todolist_template_data(
            self, slug, name, description, creator=None,
            todos="{'items': []}"):
        return {
            'slug': slug,
            'name': name,
            'description': description,
            'creator': creator,
            'todos': todos
        }

    def _verify_todolist_template_content(self, todolist_template,
                                          expected_todolist_template):
        todolist_template = dict(todolist_template)
        created_at = todolist_template.pop('created_at')
        todolist_template_id = todolist_template.pop('id')
        self.assertEqual(todolist_template, expected_todolist_template)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todolist_template_id, 0)

    def _verify_todolist_template_list(self, expected_todolist_templates):
        resp = self.request_client.get(self.todolist_template_list_url)
        self.assertEqual(resp.status_code, 200)
        data = load_encoded_json(resp.content)
        for todolist_template, expected_todolist_template in \
                zip(data, expected_todolist_templates):
            self._verify_todolist_template_content(
                todolist_template, expected_todolist_template)

    def _verify_todolist_template_creation(self):
        num_todolist_templates = TodoListTemplate.objects.all().count()
        resp = self.request_client.post(self.todolist_template_list_url, {
            'slug': self.todolist_template_slug,
            'name': self.todolist_template_name,
            'description': self.todolist_template_description})

        self.assertEqual(resp.status_code, 201)
        self.assertEqual(TodoListTemplate.objects.all().count(),
                         num_todolist_templates + 1)
        todolist_template = load_encoded_json(resp.content)
        self._verify_todolist_template_content(
            todolist_template, self._todolist_template_data(
                self.todolist_template_slug,
                self.todolist_template_name,
                self.todolist_template_description))

    def test_todolist_template_update(self):
        todolist_template = TodoListTemplateFactory()
        updated_description = 'updated description'
        todolist_template_detail_url = reverse(
            self.todolist_template_detail_url_name,
            kwargs={'pk': todolist_template.id})
        resp = self.request_client.put(
            todolist_template_detail_url,
            json.dumps(self._todolist_template_data(
                todolist_template.slug,
                todolist_template.name,
                updated_description)),
            content_type='application/json')
        updated_todolist_template = TodoListTemplateSerializer(
            TodoListTemplate.objects.get(id=todolist_template.id)).data

        self.assertEqual(resp.status_code, 200)
        self._verify_todolist_template_content(
            updated_todolist_template, self._todolist_template_data(
                todolist_template.slug,
                todolist_template.name,
                updated_description))

    def test_todolist_template_list_create(self):
        self._verify_todolist_template_list([])
        self._verify_todolist_template_creation()
        self._verify_todolist_template_list(
            [self._todolist_template_data(
                self.todolist_template_slug,
                self.todolist_template_name,
                self.todolist_template_description)])

    def _verify_todo_content(self, todo, expected_todo):
        todo = dict(todo)
        created_at = todo.pop('created_at')
        todo_id = todo.pop('id')
        todo_skipped = bool(todo.pop('skipped_datetime', None))
        expected_skipped = bool(expected_todo.pop('skipped_datetime', None))

        self.assertEqual(todo_skipped, expected_skipped)
        self.assertEqual(todo, expected_todo)
        self.assertGreater(len(created_at), 0)
        self.assertGreaterEqual(todo_id, 0)

    def test_update_todos_from_todolist_template_success(self):
        num_todos = Todo.objects.all().count()
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': [{
                'id': 1,
                'description': 'todo parent',
                'items': [{
                    'id': 2,
                    'description': 'todo child',
                    'items': []
                }]
            }]},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'task': self.task.id,
            })

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(Todo.objects.all().count(), num_todos + 3)
        todos = load_encoded_json(resp.content)
        expected_todos = [
            _todo_data(self.task, 'todo child', False,
                       template=todolist_template.id,
                       parent_todo=todos[1]['id']),
            _todo_data(self.task, 'todo parent', False,
                       template=todolist_template.id,
                       parent_todo=todos[2]['id']),
            _todo_data(self.task, self.todolist_template_name,
                       False, template=todolist_template.id),
        ]
        for todo, expected_todo in zip(todos, expected_todos):
            self._verify_todo_content(todo, expected_todo)

    def test_update_todos_from_todolist_template_missing_task_id(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': []},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug
            })

        self.assertEqual(resp.status_code, 403)

    def test_update_todos_from_todolist_template_invalid_task(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            todos={'items': []},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'task': 999999999999999,
            })

        self.assertEqual(resp.status_code, 403)

    def test_update_todos_from_todolist_template_invalid_todolist_template(
            self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': 'invalid-slug',
                'task': self.task.id,
            })

        self.assertEqual(resp.status_code, 400)

    def test_conditional_skip_remove_todos_from_template(self):
        update_todos_from_todolist_template_url = \
            reverse('orchestra:todos:update_todos_from_todolist_template')

        todolist_template = TodoListTemplateFactory(
            slug=self.todolist_template_slug,
            name=self.todolist_template_name,
            description=self.todolist_template_description,
            conditional_property_function={
                'path': 'orchestra.tests.test_todos'
                        '._get_test_conditional_props'
            },
            todos={'items': [
                {
                    'id': 1,
                    'description': 'todo parent 1',
                    'items': [{
                        'id': 2,
                        'description': 'todo child 1',
                        'items': []
                    }],
                    'remove_if': [{
                        'prop1': {
                            'operator': '==',
                            'value': True
                        }
                    }]
                }, {
                    'id': 3,
                    'description': 'todo parent 2',
                    'items': [{
                        'id': 4,
                        'description': 'todo child 2',
                        'items': [],
                        'skip_if': [{
                            'prop2': {
                                'operator': '!=',
                                'value': True
                            }
                        }]
                    }]
                }]},
        )
        resp = self.request_client.post(
            update_todos_from_todolist_template_url,
            {
                'todolist_template': todolist_template.slug,
                'task': self.task.id,
            })
        self.assertEqual(resp.status_code, 200)
        todos = load_encoded_json(resp.content)

        expected_todos = [
            _todo_data(self.task, 'todo child 2', False,
                       template=todolist_template.id,
                       parent_todo=todos[1]['id'],
                       skipped_datetime=timezone.now()),
            _todo_data(self.task, 'todo parent 2', False,
                       template=todolist_template.id,
                       parent_todo=todos[2]['id']),
            _todo_data(self.task, self.todolist_template_name,
                       False, template=todolist_template.id),
        ]
        for todo, expected_todo in zip(todos, expected_todos):
            self._verify_todo_content(todo, expected_todo)



from unittest.mock import MagicMock
from unittest.mock import patch

from dateutil.parser import parse

from django.contrib.auth import get_user_model

from orchestra.core.errors import AssignmentPolicyError
from orchestra.core.errors import CreationPolicyError
from orchestra.core.errors import IllegalTaskSubmission
from orchestra.core.errors import ModelSaveError
from orchestra.core.errors import NoTaskAvailable
from orchestra.core.errors import ReviewPolicyError
from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import TaskStatusError
from orchestra.core.errors import WorkerCertificationError
from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.tests.helpers import OrchestraTransactionTestCase
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import TaskAssignmentFactory
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.task_lifecycle import AssignmentPolicyType
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import get_new_task_assignment
from orchestra.utils.task_lifecycle import get_next_task_status
from orchestra.utils.task_lifecycle import get_task_overview_for_worker
from orchestra.utils.task_lifecycle import is_worker_certified_for_task
from orchestra.utils.task_lifecycle import role_counter_required_for_new_task
from orchestra.utils.task_lifecycle import submit_task
from orchestra.utils.task_lifecycle import tasks_assigned_to_worker
from orchestra.utils.task_lifecycle import worker_assigned_to_rejected_task
from orchestra.utils.task_lifecycle import worker_has_reviewer_status
from orchestra.utils.task_properties import current_assignment
from orchestra.workflow.defaults import get_default_creation_policy

MOCK_CURRENT = '2018-01-17T00:00:00Z'
DEADLINE1_DATETIME = '2018-01-18T00:00:00Z'
DEADLINE2_DATETIME = '2018-01-19T00:00:00Z'


class BasicTaskLifeCycleTestCase(OrchestraTransactionTestCase):
    """
    Test modular functions in the task_lifecycle
    """

    def setUp(self):
        super().setUp()
        setup_models(self)

    def test_is_worker_certified_for_task(self):
        task = Task.objects.filter(status=Task.Status.AWAITING_PROCESSING)[0]

        for worker_certification in (
                WorkerCertification.objects.filter(worker=self.workers[0])):
            worker_certification.staffbot_enabled = False
            worker_certification.save()

        # workers[0] has a certification
        self.assertTrue(
            is_worker_certified_for_task(self.workers[0],
                                         task,
                                         WorkerCertification.Role.ENTRY_LEVEL))

        self.assertFalse(
            is_worker_certified_for_task(self.workers[0],
                                         task,
                                         WorkerCertification.Role.ENTRY_LEVEL,
                                         require_staffbot_enabled=True))

        # workers[2] has no certification
        self.assertFalse(
            is_worker_certified_for_task(self.workers[2],
                                         task,
                                         WorkerCertification.Role.ENTRY_LEVEL))

    def test_not_allowed_new_assignment(self):
        invalid_statuses = [Task.Status.PROCESSING,
                            Task.Status.REVIEWING,
                            Task.Status.POST_REVIEW_PROCESSING,
                            Task.Status.COMPLETE,
                            Task.Status.ABORTED]
        for status in invalid_statuses:
            with self.assertRaises(TaskStatusError):
                get_new_task_assignment(self.workers[2], status)

    def test_get_new_task_assignment_entry_level(self):
        # Entry-level assignment
        self.assertEqual(Task.objects
                         .filter(status=Task.Status.AWAITING_PROCESSING)
                         .count(),
                         1)

        with self.assertRaises(WorkerCertificationError):
            get_new_task_assignment(self.workers[5],
                                    Task.Status.PENDING_REVIEW)

        # assign a new task to a worker
        assignment = get_new_task_assignment(self.workers[5],
                                             Task.Status.AWAITING_PROCESSING)
        self.assertTrue(assignment is not None)

        self.assertEqual(assignment.task.status,
                         Task.Status.PROCESSING)

        # No more tasks left in AWAITING_PROCESSING
        with self.assertRaises(NoTaskAvailable):
            get_new_task_assignment(self.workers[5],
                                    Task.Status.AWAITING_PROCESSING)

        # Worker should not be served machine tasks
        workflow_version = self.workflow_versions['test_workflow_2']
        simple_machine = self.workflow_steps[
            workflow_version.slug]['simple_machine']
        project = Project.objects.create(workflow_version=workflow_version,
                                         short_description='',
                                         priority=0,
                                         task_class=0)
        Task.objects.create(project=project,
                            status=Task.Status.AWAITING_PROCESSING,
                            step=simple_machine)

        with self.assertRaises(NoTaskAvailable):
            get_new_task_assignment(self.workers[5],
                                    Task.Status.AWAITING_PROCESSING)

    def test_get_new_task_assignment_reviewer(self):
        # Reviewer assignment
        self.assertEqual(Task.objects
                         .filter(status=Task.Status.PENDING_REVIEW)
                         .count(),
                         1)

        # assign a review task to worker
        assignment = get_new_task_assignment(self.workers[7],
                                             Task.Status.PENDING_REVIEW)
        self.assertTrue(assignment is not None)
        self.assertEqual(assignment.task.status,
                         Task.Status.REVIEWING)

        self.assertEqual(assignment.in_progress_task_data,
                         {'test_key': 'test_value'})

        # No tasks in state PENDING_REVIEW
        # No more tasks left in AWAITING_PROCESSING
        with self.assertRaises(NoTaskAvailable):
            get_new_task_assignment(self.workers[7],
                                    Task.Status.PENDING_REVIEW)

        # Assign an entry-level task to reviewer
        assignment = get_new_task_assignment(self.workers[7],
                                             Task.Status.AWAITING_PROCESSING)
        with self.assertRaises(NoTaskAvailable):
            get_new_task_assignment(self.workers[7],
                                    Task.Status.AWAITING_PROCESSING)

    def test_is_worker_assigned(self):
        task = self.tasks['review_task']

        # worker is not related to any task
        self.assertFalse(task.is_worker_assigned(self.workers[2]))

        # worker is assigned to a task.
        self.assertTrue(task.is_worker_assigned(self.workers[0]))

    # TODO(jrbotros): write this test when per-user max tasks logic created
    def test_worker_assigned_to_max_tasks(self):
        pass

    def test_worker_assigned_to_rejected_task(self):
        assignments = TaskAssignment.objects.filter(
            worker=self.workers[4],
            status=TaskAssignment.Status.PROCESSING,
            task__status=Task.Status.POST_REVIEW_PROCESSING)
        self.assertTrue(assignments.exists())
        self.assertTrue(worker_assigned_to_rejected_task(self.workers[4]))
        with self.assertRaises(TaskAssignmentError):
            get_new_task_assignment(self.workers[4],
                                    Task.Status.AWAITING_PROCESSING)

    def test_worker_has_reviewer_status(self):
        self.assertFalse(worker_has_reviewer_status(self.workers[0]))
        self.assertTrue(worker_has_reviewer_status(self.workers[1]))
        self.assertFalse(worker_has_reviewer_status(self.workers[2]))
        self.assertFalse(worker_has_reviewer_status(self.workers[4]))
        self.assertTrue(worker_has_reviewer_status(self.workers[5]))
        self.assertTrue(worker_has_reviewer_status(self.workers[6]))

    def test_role_counter_required_for_new_task(self):
        task = TaskFactory(status=Task.Status.COMPLETE)
        with self.assertRaises(TaskAssignmentError):
            role_counter_required_for_new_task(task)

        project = self.projects['assignment_policy']

        # Create first task in test project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)
        # Assign initial task to worker 0
        task = project.tasks.first()
        counter = role_counter_required_for_new_task(task)
        self.assertEqual(counter, 0)

        initial_task = assign_task(self.workers[0].id,
                                   task.id)
        # Submit task; next task should be created
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=True):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[0])

            counter = role_counter_required_for_new_task(initial_task)
            self.assertEqual(counter, 1)

            initial_task = assign_task(self.workers[1].id,
                                       task.id)
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[1])
            counter = role_counter_required_for_new_task(initial_task)
            self.assertEqual(counter, 2)

    def test_assign_task(self):
        entry_task = TaskFactory(
            project=self.projects['base_test_project'],
            status=Task.Status.AWAITING_PROCESSING,
            step=self.test_step)

        # No iterations should be present for task
        self.assertEqual(
            Iteration.objects.filter(assignment__task=entry_task).count(), 0)

        # Assign entry-level task to entry-level worker
        entry_task = assign_task(self.workers[0].id, entry_task.id)
        self.assertTrue(entry_task.is_worker_assigned(self.workers[0]))
        self.assertEqual(entry_task.status, Task.Status.PROCESSING)

        self.assertEqual(entry_task.assignments.count(), 1)
        entry_assignment = entry_task.assignments.first()

        # A single iteration was created for the assignment
        self.assertEqual(entry_assignment.iterations.count(), 1)
        self.assertEqual(
            Iteration.objects.filter(assignment__task=entry_task).count(), 1)
        self.assertEqual(
            entry_assignment.iterations.first().start_datetime,
            entry_assignment.start_datetime)

        # Attempt to assign task which isn't awaiting a new assignment
        invalid = (Task.Status.PROCESSING, Task.Status.ABORTED,
                   Task.Status.REVIEWING, Task.Status.COMPLETE,
                   Task.Status.POST_REVIEW_PROCESSING)
        for status in invalid:
            invalid_status_task = Task.objects.create(
                project=self.projects['base_test_project'],
                status=status,
                step=self.test_step)

            with self.assertRaises(TaskAssignmentError):
                invalid_status_task = assign_task(
                    self.workers[0].id, invalid_status_task.id)

        # Attempt to assign review task to worker already in review hierarchy
        review_task = Task.objects.create(
            project=self.projects['base_test_project'],
            status=Task.Status.PENDING_REVIEW,
            step=self.test_step)
        test_data = {'test_assign': True}
        TaskAssignmentFactory(
            worker=self.workers[1],
            task=review_task,
            status=TaskAssignment.Status.SUBMITTED,
            in_progress_task_data=test_data)

        with self.assertRaises(TaskAssignmentError):
            assign_task(self.workers[1].id, review_task.id)
        self.assertEqual(
            current_assignment(review_task).in_progress_task_data, test_data)

        # Attempt to assign review task to worker not certified for task
        with self.assertRaises(WorkerCertificationError):
            assign_task(self.workers[2].id, review_task.id)
        self.assertEqual(
            current_assignment(review_task).in_progress_task_data, test_data)

        # Assign review task to review worker
        self.assertEqual(review_task.assignments.count(), 1)
        review_task = assign_task(self.workers[3].id, review_task.id)
        self.assertEqual(review_task.assignments.count(), 2)

        reviewer_assignment = current_assignment(review_task)
        self.assertEqual(
            reviewer_assignment.worker, self.workers[3])
        self.assertEqual(
            reviewer_assignment.in_progress_task_data, test_data)
        self.assertEqual(
            reviewer_assignment.iterations.count(), 1)
        self.assertEqual(
            reviewer_assignment.iterations.first().start_datetime,
            reviewer_assignment.start_datetime)

        self.assertEqual(
            review_task.status, Task.Status.REVIEWING)

    def test_get_task_overview_for_worker(self):
        task = self.tasks['review_task']

        with self.assertRaises(TaskAssignmentError):
            get_task_overview_for_worker(task.id, self.workers[2])

        data = get_task_overview_for_worker(task.id, self.workers[0])
        expected = {
            'project': {'details': task.project.short_description,
                        'id': task.project.id,
                        'project_data': {},
                        'team_messages_url': None},
            'workflow': {'slug': 'w1',
                         'name': 'Workflow One'},
            'workflow_version': {'slug': 'test_workflow',
                                 'name': 'The workflow'},
            'prerequisites': {},
            'step': {'slug': 'step1', 'name': 'The first step'},
            'status': 'Submitted',
            'task': {'data': {'test_key': 'test_value'},
                     'status': 'Pending Review'},
            'task_id': task.id,
            'assignment_id': task.assignments.get(worker=self.workers[0]).id,
            'is_reviewer': False,
            'is_read_only': True,
            'is_project_admin': False,
            'worker': {
                'username': self.workers[0].user.username,
                'first_name': self.workers[0].user.first_name,
                'last_name': self.workers[0].user.last_name,
            }
        }
        self.assertEqual(data, expected)

        # Superuser should have access to latest task data
        superuser = get_user_model().objects.create_superuser(
            'superuser', 'superuser@b12.io', 'test-password')
        superworker = Worker.objects.create(user=superuser)
        data = get_task_overview_for_worker(task.id, superworker)
        expected['is_project_admin'] = True
        self.assertEqual(data, expected)

    def test_task_assignment_saving(self):
        """
        Ensure that workers are required for human tasks,
        and no workers are required for machine tasks.
        """
        workflow_version = self.workflow_versions['test_workflow_2']
        simple_machine = self.workflow_steps[
            workflow_version.slug]['simple_machine']
        project = Project.objects.create(workflow_version=workflow_version,
                                         short_description='',
                                         priority=0,
                                         task_class=0)
        task = Task.objects.create(project=project,
                                   status=Task.Status.PROCESSING,
                                   step=simple_machine)

        # We expect an error because a worker
        # is being saved on a machine task.
        with self.assertRaises(ModelSaveError):
            TaskAssignment.objects.create(worker=self.workers[0],
                                          task=task,
                                          status=0,
                                          in_progress_task_data={})

        human_step = self.workflow_steps[workflow_version.slug]['step4']
        task = Task.objects.create(project=project,
                                   status=Task.Status.PROCESSING,
                                   step=human_step)

        # We expect an error because no worker
        # is being saved on a human task
        with self.assertRaises(ModelSaveError):
            TaskAssignment.objects.create(task=task,
                                          status=0,
                                          in_progress_task_data={})

    def test_illegal_get_next_task_status(self):
        task = self.tasks['awaiting_processing']
        illegal_statuses = [
            Task.Status.AWAITING_PROCESSING,
            Task.Status.PENDING_REVIEW,
            Task.Status.COMPLETE
        ]

        iteration_statuses = [
            Iteration.Status.REQUESTED_REVIEW,
            Iteration.Status.PROVIDED_REVIEW
        ]

        for status in illegal_statuses:
            for iteration_status in iteration_statuses:
                with self.assertRaises(IllegalTaskSubmission):
                    task.status = status
                    get_next_task_status(task, iteration_status)

        # Entry level-related statuses cannot be rejected
        with self.assertRaises(IllegalTaskSubmission):
            task.status = Task.Status.PROCESSING
            get_next_task_status(task, Iteration.Status.PROVIDED_REVIEW)

        with self.assertRaises(IllegalTaskSubmission):
            task.status = Task.Status.POST_REVIEW_PROCESSING
            get_next_task_status(task, Iteration.Status.PROVIDED_REVIEW)

    def test_sampled_get_next_task_status(self):
        task = self.tasks['awaiting_processing']
        step = task.step
        step.review_policy = {'policy': 'sampled_review',
                              'rate': 0.5,
                              'max_reviews': 1}
        step.save()
        task.status = Task.Status.PROCESSING
        complete_count = 0
        for i in range(0, 1000):
            next_status = get_next_task_status(
                task, Iteration.Status.REQUESTED_REVIEW)
            complete_count += next_status == Task.Status.COMPLETE
        self.assertTrue(complete_count > 400)
        self.assertTrue(complete_count < 600)

    def test_legal_get_next_task_status(self):
        task = self.tasks['awaiting_processing']
        step = task.step

        task.status = Task.Status.PROCESSING
        step.review_policy = {}
        step.save()
        with self.assertRaises(ReviewPolicyError):
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW)

        step.review_policy = {'policy': 'sampled_review',
                              'rate': 1,
                              'max_reviews': 1}
        step.save()
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.PENDING_REVIEW)

        step.review_policy = {'policy': 'sampled_review',
                              'rate': 0,
                              'max_reviews': 1}
        step.save()
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.COMPLETE)

        task.status = Task.Status.POST_REVIEW_PROCESSING
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.REVIEWING)

        task = self.tasks['review_task']
        task.status = Task.Status.REVIEWING

        step.review_policy = {'policy': 'sampled_review',
                              'rate': 1,
                              'max_reviews': 0}
        step.save()
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.COMPLETE)

        step.review_policy = {'policy': 'sampled_review',
                              'rate': 1,
                              'max_reviews': 2}
        step.save()
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.PENDING_REVIEW)

        # after max reviews done a task goes to state complete
        TaskAssignment.objects.create(
            worker=self.workers[1],
            task=task,
            status=TaskAssignment.Status.SUBMITTED,
            assignment_counter=1,
            in_progress_task_data={})
        task.save()
        step.review_policy = {'policy': 'sampled_review',
                              'rate': 1,
                              'max_reviews': 1}
        step.save()
        self.assertEqual(
            get_next_task_status(task,
                                 Iteration.Status.REQUESTED_REVIEW),
            Task.Status.COMPLETE)

    def test_preassign_workers(self):
        project = self.projects['assignment_policy']

        # Create first task in test project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)
        # Assign initial task to worker 0
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        # Submit task; next task should be created
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[0])
        self.assertEqual(project.tasks.count(), 2)
        related_task = project.tasks.exclude(id=initial_task.id).first()
        # Worker 0 not certified for related tasks, so should not have been
        # auto-assigned
        self.assertEqual(related_task.assignments.count(), 0)
        self.assertEqual(related_task.status, Task.Status.AWAITING_PROCESSING)

        # Reset project
        project.tasks.all().delete()

        # Create first task in test project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)
        # Assign initial task to worker 0
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        # Submit task; verify we use the reviewer assignment policy
        mock_preassign_workers = MagicMock(return_value=initial_task)
        patch_path = 'orchestra.utils.task_lifecycle._preassign_workers'
        with patch(patch_path, new=mock_preassign_workers):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[0])
            mock_preassign_workers.assert_called_once_with(
                initial_task, AssignmentPolicyType.REVIEWER)

        # Reset project
        project.tasks.all().delete()

        # Create first task in test project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)
        # Assign initial task to worker 4
        initial_task = assign_task(self.workers[4].id,
                                   project.tasks.first().id)
        # Submit task; next task should be created
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[4])
        self.assertEqual(project.tasks.count(), 2)
        related_task = project.tasks.exclude(id=initial_task.id).first()
        # Worker 4 is certified for related task and should have been assigned
        self.assertEqual(related_task.assignments.count(), 1)
        self.assertEqual(related_task.status, Task.Status.PROCESSING)
        self.assertTrue(
            related_task.is_worker_assigned(self.workers[4]))

    def test_malformed_assignment_policy(self):
        project = self.projects['assignment_policy']
        workflow_version = project.workflow_version
        first_step = self.workflow_steps[workflow_version.slug]['step_0']

        # Create an invalid machine step with an assignment policy
        malformed_step = StepFactory(
            workflow_version=workflow_version,
            slug='machine_step',
            is_human=False,
            assignment_policy={
                'policy_function': {
                    'entry_level': {
                        'path': ('orchestra.assignment_policies.'
                                 'previously_completed_steps'),
                        'kwargs': {
                            'related_steps': ['step_0']
                        },
                    }
                }
            },
            creation_policy=get_default_creation_policy(),
        )
        malformed_step.creation_depends_on.add(first_step)

        # Create first task in project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)

        # Assign initial task to worker 0 and mark as complete
        initial_task = assign_task(self.workers[4].id,
                                   project.tasks.first().id)
        initial_task.status = Task.Status.COMPLETE
        initial_task.save()

        # Cannot preassign machine task
        with self.assertRaises(AssignmentPolicyError):
            create_subsequent_tasks(project)

        # Reset project
        project.tasks.all().delete()

        # Machine should not be member of assignment policy
        first_step.assignment_policy = {
            'policy_function': {
                'entry_level': {
                    'path': ('orchestra.assignment_policies.'
                             'previously_completed_steps'),
                    'kwargs': {
                        'related_steps': ['machine_step']
                    },
                },
            }
        }
        first_step.save()
        with self.assertRaises(AssignmentPolicyError):
            create_subsequent_tasks(project)

    def test_malformed_creation_policy(self):
        project = self.projects['creation_policy']
        workflow_version = project.workflow_version
        first_step = self.workflow_steps[
            workflow_version.slug]['creation_policy_step_0']

        # Create an invalid machine step with an assignment policy
        malformed_step = StepFactory(
            workflow_version=workflow_version,
            slug='machine_step',
            is_human=False,
            creation_policy={},
        )
        malformed_step.creation_depends_on.add(first_step)

        # Create first task in project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)

        # Assign initial task to worker 0 and mark as complete
        initial_task = assign_task(self.workers[4].id,
                                   project.tasks.first().id)
        initial_task.status = Task.Status.COMPLETE
        initial_task.save()

        # Cannot have an invalid blob for the creation_policy
        with self.assertRaises(CreationPolicyError):
            create_subsequent_tasks(project)

    @patch('orchestra.utils.task_lifecycle.schedule_machine_tasks')
    def test_schedule_machine_tasks(self, mock_schedule):
        project = self.projects['test_human_and_machine']

        # Create first task in project
        create_subsequent_tasks(project)

        # Assign initial task to worker 0 and mark as complete
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        initial_task.status = Task.Status.COMPLETE
        initial_task.save()

        create_subsequent_tasks(project)
        self.assertEqual(mock_schedule.call_count, 1)
        self.assertEqual(mock_schedule.call_args[0][0], project)
        steps = list(project.workflow_version.steps.filter(is_human=False))
        self.assertEqual(mock_schedule.call_args[0][1], steps)

    @patch('orchestra.utils.task_lifecycle._preassign_workers')
    @patch('orchestra.utils.task_lifecycle.schedule_machine_tasks')
    def test_schedule_machine_tasks_failed(self, mock_schedule,
                                           mock_preassign):
        project = self.projects['test_human_and_machine']

        # Create first task in project
        create_subsequent_tasks(project)

        # Assign initial task to worker 0 and mark as complete
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        initial_task.status = Task.Status.COMPLETE
        initial_task.save()

        mock_preassign.side_effect = Exception
        with self.assertRaises(Exception):
            create_subsequent_tasks(project)
        mock_schedule.assert_not_called()

    def test_next_todo_with_earlier_due_time(self):
        task = self.tasks['next_todo_task']
        # create todos with different due date times and one without
        TodoFactory(
            task=task,
            description='todo1',
            due_datetime=parse(DEADLINE2_DATETIME))

        TodoFactory(
            task=task,
            description='todo2')

        tasks_assigned = tasks_assigned_to_worker(self.workers[5])

        for t in tasks_assigned:
            if t['id'] == task.id:
                next_todo_due = t['next_todo_dict'].get('due_datetime', None)

                # should select the todo with a deadline
                self.assertEqual(next_todo_due, DEADLINE2_DATETIME)
                self.assertEqual(t['should_be_active'], True)

        TodoFactory(
            task=task,
            description='todo3',
            due_datetime=parse(DEADLINE1_DATETIME))

        tasks_assigned = tasks_assigned_to_worker(self.workers[5])
        for t in tasks_assigned:
            if t['id'] == task.id:
                next_todo_due = t['next_todo_dict'].get('due_datetime', None)

                # should select the todo with earliest deadline
                self.assertEqual(next_todo_due, DEADLINE1_DATETIME)
                self.assertEqual(t['should_be_active'], True)

    @patch('orchestra.utils.task_lifecycle.timezone')
    def test_next_todo_with_earlier_start_time(self, mock_timezone):
        mock_timezone.now = MagicMock(return_value=parse(
                                        MOCK_CURRENT))
        task = self.tasks['next_todo_task']
        # create todos with different due date times and one without
        TodoFactory(
            task=task,
            description='todo1',
            start_by_datetime=parse(DEADLINE2_DATETIME))

        TodoFactory(
            task=task,
            description='todo2',
            start_by_datetime=parse(DEADLINE1_DATETIME))

        tasks_assigned = tasks_assigned_to_worker(self.workers[5])
        for t in tasks_assigned:
            if t['id'] == task.id:
                next_todo_start = t['next_todo_dict'].get(
                    'start_by_datetime', None)
                # should select the todo with a deadline
                self.assertEqual(next_todo_start, DEADLINE1_DATETIME)
                self.assertEqual(t['should_be_active'], False)



from io import StringIO

from django.core.management import call_command

from orchestra.tests.helpers import OrchestraTestCase


class LoadAllWorkflowsTestCase(OrchestraTestCase):

    def test_options(self):
        # Initial load should work
        stderr1 = StringIO()
        call_command('loadallworkflows', stderr=stderr1)
        output1 = stderr1.getvalue()
        self.assertNotIn('error', output1)

        # Now it works
        stderr2 = StringIO()
        call_command('loadallworkflows', force=True, stderr=stderr2)
        output2 = stderr2.getvalue()
        self.assertNotIn('error', output2)



import logging
import random
import re
from urllib.parse import urljoin
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)

IMAGE_FILE_REGEX = re.compile('([-\w]+\.(?:jpg|jpeg|gif|png))',
                              re.IGNORECASE)


def crawl_page(project_data, prerequisites):
    """ Picks a random image off of the passed URL."""

    result = {
        'status': 'success',
        'image': None
    }
    url = project_data.get('url')
    if not url:
        result['status'] = 'error'
        result['error_message'] = 'URL was not provided.'
        return result

    # Crawl the website for images.
    logger.info('Starting to crawl %s', url)
    images = find_images(url)
    num_images = len(images)
    logger.info('Found %s images', num_images)
    if num_images == 0:
        result['status'] = 'error'
        result['error_message'] = 'Unable to find images at the provided URL.'
        return result

    # Return a random one.
    logger.info('Picking a random one...')
    image = random.choice(list(images))
    result['image'] = image
    return result


def find_images(url):
    """ Fetches a url's HTML and extracts all image sources in an <img> tag.
    """

    images = set()

    # Fetch the content.
    headers = {
        'User-Agent': ('Mozilla/5.0 (compatible; OrchestraBot/1.0; '
                       'noreply@example.org)'),
    }
    response = requests.get(url, headers=headers)
    if response.status_code < 200 or response.status_code >= 300:
        logger.error("Couldn't fetch url {}".format(url))
        return images
    content = response.text

    # Find images in the content.
    soup = BeautifulSoup(content)
    tags = soup.find_all('img', src=IMAGE_FILE_REGEX)
    for tag in tags:
        link = tag.get('src')
        if link is None:
            continue
        if not bool(urlparse(link).netloc):
            link = urljoin(url, link)
        images.add(link)
    return images



from rest_framework import permissions

from orchestra.models import Worker
from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TodoQA


class IsAssociatedWithTodosProject(permissions.BasePermission):
    """
    Ensures that a user's worker is accoiated with the todo's project.
    """

    def has_object_permission(self, request, view, obj):
        worker = Worker.objects.get(user=request.user)
        if isinstance(obj, Todo):
            project = obj.task.project
        elif isinstance(obj, TodoQA):
            project = obj.todo.task.project
        else:
            project = None
        return (
            project and
            (worker.is_project_admin() or
             worker.assignments.filter(task__project=project).exists()))


class IsAssociatedWithProject(permissions.BasePermission):
    """
    Ensures that a user's worker is associated with the request's
    `project`.

    """

    def has_permission(self, request, view):
        worker = Worker.objects.get(user=request.user)
        if worker.is_project_admin():
            return True
        if request.method == 'GET':
            # List calls have a project ID
            project_id = request.query_params.get('project')
            return worker.assignments.filter(task__project=project_id).exists()
        elif request.method == 'POST':
            # Create calls have a task ID
            task_id = request.data.get('task')
            todo_id = request.data.get('todo')
            try:
                if task_id:
                    task = Task.objects.get(id=task_id)
                elif todo_id:
                    task = Todo.objects.get(id=todo_id).task
                else:
                    task = None
                return task and \
                    worker.assignments.filter(
                        task__project=task.project).exists()
            except (Task.DoesNotExist, Todo.DoesNotExist):
                return False
        return False


class IsAssociatedWithTask(permissions.BasePermission):
    """
    Ensures that a user's worker is associated with the request's
    `task`.

    """

    def has_permission(self, request, view):
        worker = Worker.objects.get(user=request.user)
        if worker.is_project_admin():
            return True
        if request.method == 'GET':
            task_id = request.query_params.get('task')
            return worker.assignments.filter(task=task_id).exists()
        return False



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-06-29 18:27
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0076_auto_20180615_1606'),
    ]

    operations = [
        migrations.AddField(
            model_name='todolisttemplate',
            name='conditional_property_function',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AlterField(
            model_name='todo',
            name='description',
            field=models.TextField(),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-11 15:10
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0035_create_time_entries_for_snapshots'),
    ]

    operations = [
        migrations.RemoveField(  # manually-reviewed
            model_name='taskassignment',
            name='snapshots',
        ),
    ]



from io import StringIO

from django.core.management import call_command
from django.test import modify_settings

from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.workflow import assert_test_dir_v1_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v1_not_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v2_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_v2_not_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_workflow_loaded
from orchestra.tests.helpers.workflow import \
    assert_test_dir_workflow_not_loaded  # noqa

TEST_WORKFLOWS_MODULE = 'orchestra.tests.workflows'
VALID_WORKFLOW = 'test_dir'
VERSION_1 = 'test_v1'
VERSION_2 = 'test_v2'


@modify_settings(INSTALLED_APPS={
    'append': '{}.{}'.format(TEST_WORKFLOWS_MODULE, VALID_WORKFLOW)
})
class LoadWorkflowTestCase(OrchestraTestCase):

    def test_loadworkflow(self):
        """ Ensure that a workflow directory loads correctly. """
        # Verify initial DB state.
        assert_test_dir_workflow_not_loaded(self)
        assert_test_dir_v1_not_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Loading a version should create the right database objects.
        call_command('loadworkflow', VALID_WORKFLOW, VERSION_1)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_not_loaded(self)

        # Loading a second version should create the right database objects.
        call_command('loadworkflow', VALID_WORKFLOW, VERSION_2)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_loaded(self)

        # Reloading a version without --force should fail.
        stderr = StringIO()
        call_command('loadworkflow', VALID_WORKFLOW, VERSION_1,
                     stderr=stderr)
        output = stderr.getvalue()
        error_preamble = 'An error occurred while loading the workflow'
        force_error = 'Version {} already exists'.format(VERSION_1)
        self.assertIn(error_preamble, output)
        self.assertIn(force_error, output)

        # Reloading a version with --force should succeed.
        stderr2 = StringIO()
        call_command('loadworkflow', VALID_WORKFLOW, VERSION_1, force=True,
                     stderr=stderr2)
        output2 = stderr2.getvalue()
        self.assertNotIn(error_preamble, output2)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)
        assert_test_dir_v2_loaded(self)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-10-25 08:37
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0070_todos_and_reverse_assignments'),
    ]

    operations = [
        migrations.AddField(
            model_name='workercertification',
            name='staffing_priority',
            field=models.IntegerField(default=0),
        ),
    ]



from django.db.models import Max
from django.utils import timezone
from pydoc import locate

from orchestra.core.errors import SanityBotError
from orchestra.models import Project
from orchestra.models import SanityCheck
from orchestra.models import WorkflowVersion
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.utils.project_properties import incomplete_projects


def _handle(project, sanity_check, handler):
    handler_type = handler.get('type')
    handler_message = handler.get('message')
    handler_steps = handler.get('steps')
    if (handler_type != 'slack_project_channel' or
            not handler_message or not handler_steps):
        raise SanityBotError('Invalid handler: {}'.format(handler))
    tasks = (
        task for task in project.tasks.all()
        if task.step.slug in handler_steps)
    usernames = {
        assignment.worker.formatted_slack_username()
        for task in tasks
        for assignment in task.assignments.all()
        if assignment and assignment.worker
    }
    message = '{}: {}'.format(' '.join(usernames), handler_message)
    message_experts_slack_group(
        project.slack_group_id, message)


def _filter_checks(project, checks, check_configurations):
    latest_check_creation = {
        check['check_slug']: check['max_created_at']
        for check in (SanityCheck.objects
                      .filter(project=project)
                      .values('check_slug')
                      .annotate(max_created_at=Max('created_at')))}
    for check in checks:
        max_created_at = latest_check_creation.get(check.check_slug)
        seconds = (
            check_configurations.get(check.check_slug, {})
            .get('repetition_seconds'))
        now = timezone.now()
        if (max_created_at is None or
                ((seconds is not None) and
                 ((now - max_created_at).total_seconds() > seconds))):
            yield check


def _handle_sanity_checks(project, sanity_checks, check_configurations):
    sanity_checks = _filter_checks(
        project, sanity_checks, check_configurations)
    for sanity_check in sanity_checks:
        config = check_configurations.get(sanity_check.check_slug)
        if config is None:
            raise SanityBotError(
                'No configuration for {}'.format(sanity_check.check_slug))
        handlers = config.get('handlers')
        if handlers is None:
            raise SanityBotError(
                'No handlers for {}'.format(sanity_check.check_slug))
        for handler in handlers:
            _handle(project, sanity_check, handler)
        sanity_check.handled_at = timezone.now()
        sanity_check.project = project
        sanity_check.save()


def create_and_handle_sanity_checks():
    workflow_versions = WorkflowVersion.objects.all()
    for project in incomplete_projects(
            Project.objects.filter(
                workflow_version__in=workflow_versions)):
        sanity_checks = project.workflow_version.sanity_checks
        sanity_check_path = (sanity_checks
                             .get('sanity_check_function', {})
                             .get('path'))
        check_configurations = sanity_checks.get('check_configurations')
        if sanity_check_path and check_configurations:
            sanity_check_function = locate(sanity_check_path)
            sanity_checks = sanity_check_function(project)
            _handle_sanity_checks(
                project, sanity_checks, check_configurations)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import jsonfield.fields
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0008_auto_20150521_2011'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='project_data',
            field=jsonfield.fields.JSONField(default={}),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[
                                   ('copy_pass', 'Copy Pass'), ('content_extraction', ' Content Extraction')]),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='taskassignment',
            unique_together=set([('task', 'assignment_counter')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0022_auto_20151217_1642'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskassignment',
            name='status',
            field=models.IntegerField(choices=[(0, 'Processing'), (1, 'Submitted'), (2, 'Failed')]),
        ),
    ]



from orchestra.models import Certification
from orchestra.models import SanityCheck
from orchestra.models import Step
from orchestra.models import Workflow
from orchestra.models import WorkflowVersion


# Dummy machine step function so we can test machine step scheduling
def simple_json(project_data, dependencies):
    return {'json': 'simple'}


def machine_task_function():
    pass


def check_project(project):
    return [
        SanityCheck(check_slug='frequently_repeating_check'),
        SanityCheck(check_slug='infrequently_repeating_check'),
        SanityCheck(check_slug='onetime_check'),
    ]


workflow_fixtures = [
    {
        'slug': 'w1',
        'name': 'Workflow One',
        'certifications': [
            {
                'slug': 'certification1',
                'name': 'The first certification',
                'required_certifications': []
            },
            {
                'slug': 'certification2',
                'name': 'The second certification',
                'required_certifications': ['certification1']
            },
        ],
        'versions': [
            {
                'slug': 'test_workflow',
                'name': 'The workflow',
                'description': 'A description of the workflow',
                'steps': [
                    {
                        'slug': 'step1',
                        'name': 'The first step',
                        'description': ('The longer description of the '
                                        'first step'),
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': ['certification1'],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 2,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                    {
                        'slug': 'step2',
                        'name': 'The second step',
                        'description': ('The longer description of the '
                                        'second step'),
                        'is_human': True,
                        'creation_depends_on': ['step1'],
                        'required_certifications': [
                            'certification1',
                            'certification2',
                        ],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 1,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step2',
                            'angular_directive': 'step2_directive',
                        },
                    },
                    {
                        'slug': 'step3',
                        'name': 'The third step',
                        'description': ('The longer description of the '
                                        'third step'),
                        'is_human': True,
                        'creation_depends_on': ['step2'],
                        'required_certifications': [],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 1,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step3',
                            'angular_directive': 'step3_directive',
                        },
                    },
                ],
            },
        ],
    },
    {
        'slug': 'w2',
        'name': 'Workflow 2',
        'versions': [
            {
                'slug': 'test_workflow_2',
                'name': 'The workflow 2',
                'description': 'A description of the workflow',
                'steps': [
                    {
                        'slug': 'step4',
                        'name': 'The step4 for workflow2',
                        'description': ('The longer description of the '
                                        'step4 for workflow2'),
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': [],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 0,
                            'max_reviews': 0,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step4',
                            'angular_directive': 'step4_directive',
                        },
                    },
                    {
                        'slug': 'simple_machine',
                        'name': 'Simple machine step',
                        'description': 'This task returns some JSON',
                        'is_human': False,
                        'creation_depends_on': ['step4'],
                        'execution_function': {
                            'path': ('orchestra.tests.helpers.'
                                     'workflow.simple_json')
                        },
                    },
                ],
            },
        ],
    },
    {
        'slug': 'assignment_policy',
        'name': 'Assignment Policy Workflow',
        'certifications': [
            {
                'slug': 'certification1_ap',
                'name': 'The first certification',
                'required_certifications': []
            },
            {
                'slug': 'certification2_ap',
                'name': 'The second certification',
                'required_certifications': ['certification1']
            },
        ],
        'versions': [
            {
                'slug': 'assignment_policy_workflow',
                'name': 'The workflow',
                'description': 'A description of the workflow',
                'steps': [
                    {
                        'slug': 'step_0',
                        'name': 'The first step',
                        'description': ('The longer description of the '
                                        'first step'),
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': [],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 2,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                    {
                        'slug': 'step_1',
                        'name': 'The second step',
                        'description': ('The longer description of the '
                                        'second step'),
                        'is_human': True,
                        'creation_depends_on': ['step_0'],
                        'required_certifications': ['certification2_ap'],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 1,
                        },
                        'assignment_policy': {
                            'policy_function': {
                                'entry_level': {
                                    'path': ('orchestra.assignment_policies.'
                                             'previously_completed_steps'),
                                    'kwargs': {

                                        'related_steps': ['step_0'],
                                    }
                                }
                            }
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step2',
                            'angular_directive': 'step2_directive',
                        },
                    },
                ],
            },
        ],
    },
    {
        'slug': 'staffbot_assignment_policy',
        'name': 'StaffBot Assignment Policy Workflow',
        'versions': [
            {
                'slug': 'staffbot_assignment_policy_workflow',
                'name': 'The workflow',
                'description': 'A description of the workflow',
                'steps': [
                    {
                        'slug': 'staffbot_step_0',
                        'name': 'The first step',
                        'description': ('The longer description of the '
                                        'first step'),
                        'detailed_description_function': {
                            'path': ('orchestra.tests.helpers.'
                                     'fixtures.get_detailed_description')
                        },
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': ['certification1_ap'],
                        'review_policy': {
                            'policy': 'no_review',
                        },
                        'assignment_policy': {
                            'policy_function': {
                                'entry_level': {
                                    'path': ('orchestra.bots.'
                                             'assignment_policies.'
                                             'staffbot_autoassign'),
                                }
                            }
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                    {
                        'slug': 'staffbot_step_1',
                        'name': 'The second step',
                        'description': ('The longer description of the '
                                        'second step'),
                        'detailed_description_function': {
                            'path': ('orchestra.tests.helpers.'
                                     'fixtures.get_detailed_description'),
                            'kwargs': {
                                'text': 'step 1 text',
                            },
                        },
                        'is_human': True,
                        'creation_depends_on': ['staffbot_step_0'],
                        'required_certifications': ['certification1_ap'],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 1,
                        },
                        'assignment_policy': {
                            'policy_function': {
                                'entry_level': {
                                    'path': ('orchestra.bots.'
                                             'assignment_policies.'
                                             'staffbot_autoassign'),
                                    'kwargs': {

                                        'related_steps': ['staffbot_step_0'],
                                    }
                                }
                            }
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                ],
            },
        ],
    },
    {
        'slug': 'creation_policy',
        'name': 'Creation Policy Workflow',
        'versions': [
            {
                'slug': 'creation_policy_workflow',
                'name': 'The workflow',
                'description': 'A description of the workflow',
                'steps': [
                    {
                        'slug': 'creation_policy_step_0',
                        'name': 'The first step',
                        'description': ('The longer description of the '
                                        'first step'),
                        'detailed_description_function': {
                            'path': ('orchestra.tests.helpers.'
                                     'fixtures.get_detailed_description')
                        },
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': [],
                        'review_policy': {
                            'policy': 'no_review',
                        },
                        'creation_policy': {
                            'policy_function': {
                                'path': ('orchestra.creation_policies'
                                         '.always_create'),
                            }
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                    {
                        'slug': 'creation_policy_step_1',
                        'name': 'The second step',
                        'description': ('The longer description of the '
                                        'second step'),
                        'detailed_description_function': {
                            'path': ('orchestra.tests.helpers.'
                                     'fixtures.get_detailed_description'),
                            'kwargs': {
                                'text': 'step 1 text',
                            },
                        },
                        'is_human': True,
                        'creation_depends_on': ['creation_policy_step_0'],
                        'required_certifications': ['certification1_ap'],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 1,
                        },
                        'creation_policy': {
                            'policy_function': {
                                'path': ('orchestra.utils.tests.'
                                         'test_creation_policies.never_create'
                                         )
                            }
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                ],
            },
        ],
    },

    # The following is a workflow with a more complex dependency graph to test
    # the topographical sort and other graph related functions.
    #
    # A - C \     / G
    #        E - F
    # B - D /     \ H
    #
    # Correct ordering would be:
    # [A B] [C D] E F [G H]
    # where letters in brackets can go in either order
    {
        'slug': 'w3',
        'name': 'Workflow 3',
        'versions': [
            {
                'slug': 'crazy_workflow',
                'name': 'The crazy workflow',
                'description': 'A description of the crazy workflow',
                'steps': [
                    {
                        'slug': 'stepA',
                        'name': 'Step A',
                        'description': 'Step A',
                        'is_human': True,
                        'creation_depends_on': [],
                    },
                    {
                        'slug': 'stepB',
                        'name': 'Step B',
                        'description': 'Step B',
                        'is_human': True,
                        'creation_depends_on': [],
                    },
                    {
                        'slug': 'stepC',
                        'name': 'Step C',
                        'description': 'Step C',
                        'is_human': True,
                        'creation_depends_on': ['stepA'],
                    },
                    {
                        'slug': 'stepD',
                        'name': 'Step D',
                        'description': 'Step D',
                        'is_human': True,
                        'creation_depends_on': ['stepB'],
                    },
                    {
                        'slug': 'stepE',
                        'name': 'Step E',
                        'description': 'Step E',
                        'is_human': True,
                        'creation_depends_on': ['stepC', 'stepD'],
                    },
                    {
                        'slug': 'stepF',
                        'name': 'Step F',
                        'description': 'Step F',
                        'is_human': True,
                        'creation_depends_on': ['stepE'],
                    },
                    {
                        'slug': 'stepG',
                        'name': 'Step G',
                        'description': 'Step G',
                        'is_human': True,
                        'creation_depends_on': ['stepF'],
                    },
                    {
                        'slug': 'stepH',
                        'name': 'Step H',
                        'description': 'Step H',
                        'is_human': True,
                        'creation_depends_on': ['stepF'],
                    },
                ],
            },
        ],
    },

    # The following is a workflow that contains a cycle.
    {
        'slug': 'w4',
        'name': 'Workflow 4',
        'versions': [
            {
                'slug': 'erroneous_workflow_1',
                'name': 'Erroneous Workflow 1',
                'description': 'The workflow is wrong',
                'steps': [
                    {
                        'slug': 'stepA',
                        'name': 'Step A',
                        'description': 'Step A',
                        'is_human': True,
                        'creation_depends_on': [],
                    },
                    {
                        'slug': 'stepB',
                        'name': 'Step B',
                        'description': 'Step B',
                        'is_human': True,
                        'creation_depends_on': ['stepA', 'stepC'],
                    },
                    {
                        'slug': 'stepC',
                        'name': 'Step C',
                        'description': 'Step C',
                        'is_human': True,
                        'creation_depends_on': ['stepB'],
                    },
                ],
            },
        ],
    },

    # The following is a workflow that has no starting point (and also a
    # cycle).
    {
        'slug': 'w5',
        'name': 'Workflow 5',
        'versions': [
            {
                'slug': 'erroneous_workflow_2',
                'name': 'Erroneous Workflow 2',
                'description': 'The workflow is wrong',
                'steps': [
                    {
                        'slug': 'stepA',
                        'name': 'Step A',
                        'description': 'Step A',
                        'is_human': True,
                        'creation_depends_on': ['stepB'],
                    },
                    {
                        'slug': 'stepB',
                        'name': 'Step B',
                        'description': 'Step B',
                        'is_human': True,
                        'creation_depends_on': ['stepA'],
                    },
                ],
            },
        ],
    },

    # The following is a workflow with a single machine step.
    {
        'slug': 'machine_workflow',
        'name': 'Machine workflow',
        'versions': [
            {
                'slug': 'machine_workflow_version',
                'name': 'v1',
                'description': 'A workflow with a single machine step',
                'steps': [
                    {
                        'slug': 'machine_step',
                        'name': 'Machine step',
                        'description': 'A step to be carried out by a machine',
                        'is_human': False,
                        'creation_depends_on': [],
                        'required_certifications': [],
                        'execution_function': {},
                    },
                ],
            },
        ],
    },
    {
        'slug': 'sanitybot',
        'name': 'SanityBot workflow',
        'certifications': [],
        'versions': [
            {
                'slug': 'sanitybot_workflow',
                'name': 'SanityBot workflow',
                'description': 'A description of the SanityBot workflow',
                'sanity_checks': {
                    'sanity_check_function': {
                        'path': ('orchestra.tests.helpers.workflow'
                                 '.check_project')
                    },
                    'check_configurations': {
                        "frequently_repeating_check": {
                            "handlers": [{
                                "type": "slack_project_channel",
                                "message": "Frequently repeating message",
                                "steps": ["step1"]
                            }],
                            "repetition_seconds": 86400  # One day
                        },
                        "infrequently_repeating_check": {
                            "handlers": [{
                                "type": "slack_project_channel",
                                "message": "Infrequently repeating message",
                                "steps": ["step1"]
                            }],
                            "repetition_seconds": 864000  # Ten days
                        },
                        "onetime_check": {
                            "handlers": [{
                                "type": "slack_project_channel",
                                "message": "One-time message",
                                "steps": ["step1"]
                            }],
                        },
                        "unused_check": {
                            "handlers": [{
                                "type": "slack_project_channel",
                                "message": "Unused message",
                                "steps": ["step1"]
                            }],
                        },
                    }
                },
                'steps': [
                    {
                        'slug': 'step1',
                        'name': 'The first step',
                        'description': ('The longer description of the '
                                        'first step'),
                        'is_human': True,
                        'creation_depends_on': [],
                        'required_certifications': [],
                        'review_policy': {
                            'policy': 'sampled_review',
                            'rate': 1,
                            'max_reviews': 2,
                        },
                        'user_interface': {
                            'javascript_includes': ['/path/to/some.js'],
                            'stylesheet_includes': ['/path/to/some.css'],
                            'angular_module': 'step1',
                            'angular_directive': 'step1_directive',
                        },
                    },
                ],
            },
        ],
    },
]


def assert_test_dir_workflow_not_loaded(test_case):
    """ Verify that the test_dir workflow's DB objects don't exist. """
    test_case.assertModelInstanceNotExists(Workflow, {'slug': 'test_dir'})
    test_case.assertModelInstanceNotExists(Certification, {
        'slug': 'certification1',
        'workflow__slug': 'test_dir',
    })
    test_case.assertModelInstanceNotExists(Certification, {
        'slug': 'certification2',
        'workflow__slug': 'test_dir',
    })


def assert_test_dir_workflow_loaded(test_case):
    """ Verify that the test_dir workflow's DB objects exist. """
    workflow = test_case.assertModelInstanceExists(
        Workflow,
        {
            'slug': 'test_dir',
        },
        {
            'name': 'Test Workflow',
            'description': 'Test workflow functionality.',
        })

    certification1 = test_case.assertModelInstanceExists(
        Certification,
        {
            'workflow': workflow,
            'slug': 'certification1',
        },
        {
            'name': 'First Certification',
            'description': 'The first one',
        })
    test_case.assertEqual(list(certification1.required_certifications.all()),
                          [])

    certification2 = test_case.assertModelInstanceExists(
        Certification,
        {
            'workflow': workflow,
            'slug': 'certification2',
        },
        {
            'name': 'Second Certification',
            'description': 'The second one',
        })
    test_case.assertEqual(list(certification2.required_certifications.all()),
                          [certification1])


def assert_test_dir_v1_not_loaded(test_case):
    """ Verify that the test_dir workflow v1's DB objects don't exist. """
    test_case.assertModelInstanceNotExists(WorkflowVersion, {
        'slug': 'test_v1',
        'workflow__slug': 'test_dir',
    })
    test_case.assertModelInstanceNotExists(Step, {
        'slug': 's1',
        'workflow_version__slug': 'test_v1',
    })
    test_case.assertModelInstanceNotExists(Step, {
        'slug': 's2',
        'workflow_version__slug': 'test_v1',
    })
    test_case.assertModelInstanceNotExists(Step, {
        'slug': 's3',
        'workflow_version__slug': 'test_v1',
    })


def assert_test_dir_v1_loaded(test_case):
    """ Verify that the test_dir workflow v1's DB objects exist. """
    version1 = test_case.assertModelInstanceExists(
        WorkflowVersion,
        {
            'workflow__slug': 'test_dir',
            'slug': 'test_v1',
        },
        {
            'name': 'Test Workflow v1',
            'description': 'First test workflow version',
        })

    step1 = test_case.assertModelInstanceExists(
        Step,
        {
            'workflow_version': version1,
            'slug': 's1',
        },
        {
            'name': 'Step 1',
            'description': 'The first step',
            'detailed_description_function': {
                'path': ('orchestra.tests.helpers.'
                         'fixtures.get_detailed_description'),
                'kwargs': {
                    'text': 'step 2 text'
                }
            },
            'is_human': False,
            'execution_function': {
                'path': 'v1.machine.machine_function',
            },
            'review_policy': {},
            'user_interface': {},
        })
    test_case.assertEqual(list(step1.required_certifications.all()), [])
    test_case.assertEqual(list(step1.creation_depends_on.all()), [])
    test_case.assertEqual(list(step1.submission_depends_on.all()), [])

    step2 = test_case.assertModelInstanceExists(
        Step,
        {
            'workflow_version': version1,
            'slug': 's2',
        },
        {
            'name': 'Step 2',
            'description': 'The second step',
            'is_human': True,
            'execution_function': {},
            'review_policy': {
                'policy': 'no_review',
            },
            'user_interface': {
                'angular_module': 'test_dir.v1.s2',
                'angular_directive': 's2',
                'javascript_includes': [
                    'test_dir/v1/s2/js/modules.js',
                    'test_dir/v1/s2/js/controllers.js',
                    'test_dir/v1/s2/js/directives.js',
                ],
            },
        })
    test_case.assertEqual(list(step2.required_certifications.all()),
                          [Certification.objects.get(
                              slug='certification1',
                              workflow__slug='test_dir',
                          )])
    test_case.assertEqual(list(step2.creation_depends_on.all()), [step1])
    test_case.assertEqual(list(step2.submission_depends_on.all()), [])

    step3 = test_case.assertModelInstanceExists(
        Step,
        {
            'workflow_version': version1,
            'slug': 's3',
        },
        {
            'name': 'Step 3',
            'description': 'The third step',
            'is_human': True,
            'execution_function': {},
            'review_policy': {
                'policy': 'sampled_review',
                'rate': 1,
                'max_reviews': 1,
            },
            'user_interface': {
                'angular_module': 'test_dir.v1.s3',
                'angular_directive': 's3',
                'javascript_includes': [
                    'test_dir/v1/s3/js/modules.js',
                    'test_dir/v1/s3/js/controllers.js',
                    'test_dir/v1/s3/js/directives.js',
                ],
            },
        })
    test_case.assertEqual(set(step3.required_certifications.all()),
                          set(Certification.objects.filter(
                              slug__in=['certification1', 'certification2'],
                              workflow__slug='test_dir',
                          )))
    test_case.assertEqual(list(step3.creation_depends_on.all()), [step1])
    test_case.assertEqual(list(step3.submission_depends_on.all()), [step2])


def assert_test_dir_v2_not_loaded(test_case):
    """ Verify that the test_dir workflow v2's DB objects exist. """
    test_case.assertModelInstanceNotExists(WorkflowVersion, {
        'slug': 'test_v2',
        'workflow__slug': 'test_dir',
    })
    test_case.assertModelInstanceNotExists(Step, {
        'slug': 's1',
        'workflow_version__slug': 'test_v2',
    })
    test_case.assertModelInstanceNotExists(Step, {
        'slug': 's2',
        'workflow_version__slug': 'test_v2',
    })


def assert_test_dir_v2_loaded(test_case):
    """ Verify that the test_dir workflow v2's DB objects exist. """
    version2 = test_case.assertModelInstanceExists(
        WorkflowVersion,
        {
            'workflow__slug': 'test_dir',
            'slug': 'test_v2',
        },
        {
            'name': 'Test Workflow V2',
            'description': 'Second test workflow version',
        })

    step1 = test_case.assertModelInstanceExists(
        Step,
        {
            'workflow_version': version2,
            'slug': 's1',
        },
        {
            'name': 'Step 1',
            'description': 'The first step.',
            'is_human': False,
            'execution_function': {
                'path': 'v2.machine.machine_function'
            },
            'review_policy': {},
            'user_interface': {},
        })
    test_case.assertEqual(list(step1.required_certifications.all()), [])
    test_case.assertEqual(list(step1.creation_depends_on.all()), [])
    test_case.assertEqual(list(step1.submission_depends_on.all()), [])

    step2 = test_case.assertModelInstanceExists(
        Step,
        {
            'workflow_version': version2,
            'slug': 's2',
        },
        {
            'name': 'Step 2',
            'description': 'The second step',
            'is_human': True,
            'execution_function': {},
            'review_policy': {
                'policy': 'no_review',
            },
            'user_interface': {
                'angular_module': 'test_dir.v2.s2',
                'angular_directive': 's2',
                'javascript_includes': [
                    'test_dir/v2/s2/js/modules.js',
                    'test_dir/v2/s2/js/controllers.js',
                    'test_dir/v2/s2/js/directives.js',
                ],
            },
        })
    test_case.assertEqual(list(step2.required_certifications.all()),
                          [Certification.objects.get(
                              slug='certification2',
                              workflow__slug='test_dir',
                          )])
    test_case.assertEqual(list(step2.creation_depends_on.all()), [step1])
    test_case.assertEqual(list(step2.submission_depends_on.all()), [])



from unittest.mock import patch

from orchestra.models import Iteration
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import submit_task


def never_create(prerequisite_data, project_data, **kwargs):
    return False


class CreationPolicyTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)

    def test_always_create_policy(self):
        project = self.projects['creation_policy']

        # Create first task in test project
        create_subsequent_tasks(project)
        self.assertEqual(project.tasks.count(), 1)

        # Assign initial task to worker 0
        initial_task = assign_task(self.workers[0].id,
                                   project.tasks.first().id)
        # Submit task; next task should not be created, it never is
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[0])

        self.assertEqual(project.tasks.count(), 1)



from django.conf import settings
from django.core.mail import send_mail as _send_mail
from markdown2 import markdown

from orchestra.models import CommunicationPreference
from orchestra.models import Worker


def send_mail(subject, message, from_email,
              recipient_list, fail_silently=False,
              auth_user=None, auth_password=None,
              connection=None, html_message=None,
              communication_type=None):
    """
    Light wrapper over Django's send_mail which filters out recipients who
    opted out of using email via their `CommuncationPreference` object.
    If `mock_mail` is set to true, the email is sent to the `from_email` to
    allow for testing.
    """
    if communication_type is not None:
        recipient_list = [
            email for email in recipient_list
            if _can_email(communication_type, email)
        ]
    if settings.ORCHESTRA_MOCK_EMAILS:
        # Send the mail to as many people as we normally would
        recipient_list = [
            settings.ORCHESTRA_MOCK_TO_EMAIL] * len(recipient_list)
    if len(recipient_list):
        return _send_mail(subject, message, from_email,
                          recipient_list, fail_silently=fail_silently,
                          auth_user=auth_user, auth_password=auth_password,
                          connection=connection, html_message=html_message)


def html_from_plaintext(plaintext):
    """
    Convert a plaintext message to html by rendering it as markdown.  We add
    extra newlines so that markdown inserts the appropriate <p> tags
    """
    return markdown(plaintext.replace('\n', '\n\n'))


def _can_email(communication_type, email):
    """
        Try to get the Worker associated with the given email and query if they
        want to receive a message for the given communication_type.
    """
    try:
        worker = Worker.objects.get(user__email=email)
        comm_pref = CommunicationPreference.objects.get(
            worker=worker,
            communication_type=communication_type
        )
        return comm_pref.can_email()
    except Worker.DoesNotExist:
        return True
